/***********************************************************************/
/*                                                                     */
/* PIDL.cpp    : Source file for ITEMIDLIST wrapper class              */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/PIDL.cpp $                   */
/*  $Date: 2005-11-14 20:30:56 +0900 (Mon, 14 Nov 2005) $              */
/*   $Rev: 54 $                                                        */
/*                                                                     */
/***********************************************************************/

#include <algorithm>
#include <iostream>

#include <shlwapi.h>

#include "PIDL.hpp"
#include "odstream.hpp"

using yak::debug::ods;

namespace yak {
	namespace utils {
		namespace windows {

			PIDL::PIDL(const PIDL& r):pidl(NULL)
			{
				ods << "PIDL::PIDL(" << &r << ')' << std::endl;
				UINT cb = r.getSize();
				if(cb) {
					pidl = (LPITEMIDLIST)CoTaskMemAlloc(cb);
					ods << "PIDL::PIDL(const PIDL&) - CoTaskMemAlloc():" << pidl << std::endl;
					if(pidl) {
						CopyMemory(pidl, r.pidl, cb);
					}
					// TODO: should throw exception
				}
			}

			PIDL& PIDL::operator=(const PIDL& r)
			{
				ods << "PIDL::operator=(" << &r << ')' << std::endl;
				if(pidl == r.pidl) return *this;
				PIDL temp(r);
				return swap(temp);
			}

			PIDL& PIDL::swap(PIDL& r)
			{
				std::swap(pidl, r.pidl);
				return *this;
			}

			PIDL::~PIDL(void)
			{
				ods << "PIDL::~PIDL()" << std::endl;
				ods << "PIDL::~PIDL() - CoTaskMemFree():" << pidl << std::endl;
				CoTaskMemFree(pidl);
			}

			PIDL::operator LPITEMIDLIST* (void)
			{
				reset();
				return &pidl;
			}

			PIDL& PIDL::assign(LPITEMIDLIST pidl_)
			{
				ods << "PIDL::assign(" << pidl_ << ')' << pidl << std::endl;
				reset();
				pidl = pidl_;
				return *this;
			}

			// from SDK example UINT GetSize(LPCITEMIDLIST pidl)
			UINT PIDL::getSize(void) const
			{
			    UINT cbTotal = 0;
			    LPITEMIDLIST pidl_ = pidl;
			    if (pidl_)
			    {
			        cbTotal += sizeof(pidl_->mkid.cb);    // Terminating null character
			        while (pidl_)
			        {
			            cbTotal += pidl_->mkid.cb;
			            pidl_ = getNextItemID(pidl_);
			        }
			    }
				return cbTotal;
			}

			PIDL& PIDL::operator += (const PIDL& pidl_)
			{
				PIDL temp(*this + pidl_);
				return swap(temp);
			}

			// from SDK example BOOL getParent(LPITEMIDLIST pidl)
			PIDL PIDL::getParent(void) const
			{
				if(!pidl) return *this; // TODO: should throw

				PIDL temp(*this);
				LPITEMIDLIST pidl_ = temp.pidl;
				if(pidl_->mkid.cb) {
					LPITEMIDLIST pidlNext = pidl_;
					while(pidlNext) {
						pidl_ = pidlNext;
						pidlNext = getNextItemID(pidl_);
					}
					pidl_->mkid.cb = 0;
					return temp;
				}
				return *this; // TODO: should throw
			}

			// from SDK example LPITEMIDLIST Append(LPCITEMIDLIST pidlBase, LPCITEMIDLIST pidlAdd)
			PIDL operator + (const PIDL& r1, const PIDL& r2)
			{
				ods << "PIDL operator +(" << &r1 << ',' << &r2 << ')' << std::endl;
				if(r1.getSize() == 0) return r2;
				UINT cb1 = r1.getSize() - sizeof(r1.pidl->mkid.cb); // Terminating null character
				UINT cb2 = r2.getSize();
				LPITEMIDLIST pidlNew = (LPITEMIDLIST)CoTaskMemAlloc(cb1 + cb2);
				ods << "PIDL operator +() - CoTaskMemAlloc(): " << pidlNew << std::endl;
				if(pidlNew) {
					CopyMemory(pidlNew, r1.pidl, cb1);
					CopyMemory(((LPSTR)pidlNew) + cb1, r2.pidl, cb2);
					return PIDL().assign(pidlNew);
				}
				// TODO: should throw exception
				return PIDL();
			}

			// from SDK example LPITEMIDLIST GetNextItemID(LPCITEMIDLIST pidl)
			LPITEMIDLIST PIDL::getNextItemID(LPCITEMIDLIST pidl)
			{
			   // Check for valid pidl.
			   if(pidl == NULL)
			      return NULL;

			   // Get the size of the specified item identifier. 
			   int cb = pidl->mkid.cb; 

			   // If the size is zero, it is the end of the list. 
			   if (cb == 0) 
			      return NULL; 

			   // Add cb to pidl (casting to increment by bytes). 
			   pidl = (LPITEMIDLIST) (((LPBYTE) pidl) + cb); 

			   // Return NULL if it is null-terminating, or a pidl otherwise. 
			   return (pidl->mkid.cb == 0) ? NULL : (LPITEMIDLIST) pidl; 
			}

			void PIDL::reset(void)
			{
				ods << "PIDL::reset()" << std::endl;
				CoTaskMemFree(pidl);
				ods << "PIDL::reset() - CoTaskMemFree():" << pidl << std::endl;
				pidl = NULL;
			}

			std::ostream& operator<< (std::ostream &os, const PIDL &r)
			{
				LPITEMIDLIST pidl = r.pidl;
				if(pidl == NULL) return os;
				int cb = pidl->mkid.cb;
				while(cb != 0) {
					for(int i=0;i<cb;i++) os << (int)(pidl->mkid.abID[i]) << ',';
					os << std::endl;
					pidl = PIDL::getNextItemID(pidl);
					if(pidl) {
						cb = pidl->mkid.cb;
					} else {
						cb = 0;
					}
				}
				return os;
			}

		} // namespace windows
	} // namespace utils
} // namespace yak
