/***********************************************************************/
/*                                                                     */
/* Slurp.cpp   : Source file for mmap utility class                    */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/Slurp.cpp $                  */
/*  $Date: 2005-11-12 17:45:03 +0900 (Sat, 12 Nov 2005) $              */
/*   $Rev: 55 $                                                        */
/*                                                                     */
/***********************************************************************/

#include <iostream>

#include <windows.h>

#include "Slurp.hpp"
#include "odstream.hpp"

using yak::debug::ods;

namespace yak {
	namespace utils {
		namespace windows {

			SlurpBase::SlurpBase(LPSTR filename, DWORD dwOffset):hFile(INVALID_HANDLE_VALUE), hMap(INVALID_HANDLE_VALUE), pbActualStart(NULL), pbStart(NULL), dwSize(0)
			{
				ods << "SlurpBase::SlurpBase(LPSTR)" << std::endl;

				hFile = ::CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
				if (hFile == INVALID_HANDLE_VALUE) {
					ods << "SlurpBase::SlurpBase(LPSTR): Failed at CreateFile()" << std::endl;
					return; // TODO: should throw
				}

				init(hFile, dwOffset);
			}

			SlurpBase::SlurpBase(HANDLE hFile_, DWORD dwOffset):hFile(INVALID_HANDLE_VALUE), hMap(INVALID_HANDLE_VALUE), pbActualStart(NULL), pbStart(NULL), dwSize(0)
			{
				init(hFile_, dwOffset);
			}

			void SlurpBase::init(HANDLE hFile_, DWORD dwOffset)
			{
				dwSize = ::GetFileSize(hFile, NULL);
				ods << "SlurpBase::init() : dwSize = " << dwSize << std::endl;
				if(dwSize < dwOffset) {
					ods << "SlurpBase::init() : dwSize = " << dwSize << " < dwOffset = " << dwOffset << std::endl;
					return; // TODO: should throw
				}

				hMap = ::CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, 0, NULL);
				if (hMap == NULL) {
					ods << "SlurpBase::init(): Failed at CreateFileMapping()" << std::endl;
					::CloseHandle(hFile);
					return; // TODO: should throw
				}

				pbActualStart = (LPBYTE)MapViewOfFile(hMap, FILE_MAP_READ, 0, 0, 0);
				if (pbActualStart == NULL) {
					ods << "SlurpBase::init(): Failed at MapViewOfFile()" << std::endl;
					::CloseHandle(hMap);
					::CloseHandle(hFile);
					return; // TODO: should throw
				}

				dwSize -= dwOffset;
				pbStart = pbActualStart + dwOffset;
			}

			SlurpBase::~SlurpBase()
			{
				ods << "SlrupBase::~SlurpBase()" << std::endl;
				::UnmapViewOfFile(pbActualStart);
				::CloseHandle(hMap);
				::CloseHandle(hFile);
			}

		} // namespace windows
	} // namespace utils
} // namespace yak
