/***********************************************************************/
/*                                                                     */
/* afxspdir.cpp: Source file for SPI to open special folder in Afx     */
/*                                                                     */
/*               Copyright (c) Yak! 2005 All rights reserved.          */
/*                                                                     */
/*   $URL: svn://localhost/afxspdir/trunk/afxspdir.cpp $               */
/*  $Date: 2005-11-14 20:30:56 +0900 (Mon, 14 Nov 2005) $              */
/*   $Rev: 54 $                                                        */
/*                                                                     */
/***********************************************************************/

#include <iostream>
#include <iomanip>
#include <string>
#include <algorithm>
#include <cstring> // for memcmp

#include <boost/lexical_cast.hpp>

#include <windows.h>

#include "Spi_api.h"
#include "Slurp.hpp"
#include "parseini.hpp"
#include "infoman.hpp"

#ifndef NDEBUG
#define DEBUG
#endif
#include "odstream.hpp"

using yak::debug::ods;
using boost::lexical_cast;

// TODO: Use singleton
InfoManager g_imSpdir(5);

// from spdir.cpp
extern void MakeFileInfo(LPOLESTR pCLSID, const std::string &sKey, int depth, int max_entry, int max_entry_indir);
extern HANDLE CreateLinkOnMemory(LPITEMIDLIST pidl, LPSTR lpszDesc);

// Force null terminating version of strncpy
// Return length without null terminator
int safe_strncpy(char *dest, const char *src, std::size_t n)
{
	size_t i = 0;
	while(i<n && *src) { *dest++ = *src++; ++i; }
	*dest = 0;
	return i;
}

const char* table[] = {
	"00AM",
	"Bridge plugin for special folder in Afx - v0.01 (c) Yak!",
	"*.ini",
	"ݒt@C"
};

INT PASCAL GetPluginInfo(INT infono, LPSTR buf, INT buflen)
{
	ods << "GetPluginInfo(" << infono << ',' << buf << ',' << buflen << ')' << std::endl;
	if(0 <= infono && static_cast<size_t>(infono) < sizeof(table)/sizeof(table[0])) {
		return safe_strncpy(buf, table[infono], buflen);
	} else {
		return 0;
	}
}

static INT IsSupportedImp(LPSTR filename, LPBYTE pb)
{
	if(memcmp(pb, "[afxspdir]", 10) == 0) return SPI_SUPPORT_YES;
	return SPI_SUPPORT_NO;
}

INT PASCAL IsSupported(LPSTR filename, DWORD dw)
{
	ods << "IsSupported(" << filename << ',' << std::hex << std::setw(8) << std::setfill('0') << dw << ')' << std::endl;
	if(HIWORD(dw) == 0) {
		ods << "File handle" << std::endl;
		BYTE pb[2048];
		DWORD dwSize;
		ReadFile((HANDLE)dw, pb, sizeof(pb), &dwSize, NULL);
		return IsSupportedImp(filename, pb);;
	} else {
		ods << "Pointer" << std::endl; // By afx
		return IsSupportedImp(filename, (LPBYTE)dw);;
	}
	// not reached here
}

static void SetArchiveInfo(const std::string &sKey, LPSTR buf, DWORD len)
{
	ParseINI piConf(std::make_pair(buf, buf + len));
ods << "SetArchiveInfoImp - ParseINI is initialized" << std::endl;
	WCHAR wsz[MAX_PATH];

ods << "Type: " << piConf.getType() << " clsid_path: " << piConf.getValue("clsid_path") << std::endl;
	// Ensure that the string is Unicode. 
	int nDepth = piConf.isExist("max_depth") ? lexical_cast<int>(piConf.getValue("max_depth")) : 3;
	int nEntry = piConf.isExist("max_entry") ? lexical_cast<int>(piConf.getValue("max_entry")) : 100;
	int nEntryInDir = piConf.isExist("max_entry_indir") ? lexical_cast<int>(piConf.getValue("max_entry_indir")) : 30;
	if(piConf.getValue("clsid_path") == "::") {
		MakeFileInfo(NULL, sKey, nDepth, nEntry, nEntryInDir);
	} else {
		MultiByteToWideChar(CP_ACP, 0, piConf.getValue("clsid_path").c_str(), -1, wsz, MAX_PATH); 
		// TODO: Check return value from MultiByteWideChar to ensure success.
		MakeFileInfo(wsz, sKey, nDepth, nEntry, nEntryInDir);
	}
ods << "SetArchiveInfoImp - Making File info ends" << std::endl;
}

static INT GetArchiveInfoImp(LPSTR buf, DWORD len, HLOCAL *lphInf, LPSTR filename = NULL)
{
ods << "GetArchiveInfoImp(" << std::string(buf, std::min<DWORD>(len, 1024)) << ',' << len << ',' << lphInf << (filename ? filename : "NULL") << ')' << std::endl;
	static int nUnique;
	std::string sKey = filename ? filename : "::asd::" + lexical_cast<std::string>(nUnique++);
	if(filename == NULL || !g_imSpdir.isCached(sKey)) { // Cache mishit
ods << "GetArchiveInfoImp - Cache mishit by " << sKey << std::endl;
		SetArchiveInfo(sKey, buf, len);
	}
	std::size_t size = g_imSpdir.getFileInfoSize(sKey);
ods << "GetArchiveInfoImp - Fileinfo size is " << size << std::endl;
	size += sizeof(SPI_FILEINFO); // for terminator
	*lphInf = LocalAlloc(LMEM_MOVEABLE, size);
	LPVOID pv = LocalLock(*lphInf);
	ZeroMemory(pv, size);
	g_imSpdir.copyFileInfo(filename, pv, size);
	LocalUnlock(*lphInf);
	return SPI_ERR_NO_ERROR;
}

INT PASCAL GetArchiveInfo(LPSTR buf, LONG len, UINT flag, HLOCAL *lphInf)
{
	ods << "GetArchiveInfo(" << std::string(buf, std::min<DWORD>(len, 1024)) << ',' << len << ',' << std::hex << std::setw(8) << std::setfill('0') << flag << ',' << lphInf << ')' << std::endl;
	switch(flag & 7) {
	case 0:
	  {
		ods << "File" << std::endl; // By afx
		yak::utils::windows::Slurp<CHAR> s(buf, len);
		return GetArchiveInfoImp(s.begin(), s.getSize(), lphInf, buf);
	  }
	case 1:
		ods << "Memory" << std::endl;
		return GetArchiveInfoImp(buf, len, lphInf);
	}
	return SPI_ERR_NO_ERROR;
}

INT PASCAL GetFileInfo(LPSTR buf, LONG len, LPSTR filename, UINT flag, SPI_FILEINFO *lpInfo)
{
	ods << "GetFileInfo(" << std::string(buf, std::min<DWORD>(len, 1024)) << ',' << len << ',' << filename << ','<< std::hex << std::setw(8) << std::setfill('0') << flag << ',' << lpInfo << ')' << std::endl;
	if(flag & 128) {
		ods << "Case-insensitive" << std::endl; // By afx
	} else {
		ods << "Case-sensitive" << std::endl;
	}
	switch(flag & 7) {
	case 0:
		ods << "File" << std::endl; // By afx
		break;
	case 1:
		ods << "Memory" << std::endl;
		return SPI_ERR_NOT_IMPLEMENTED;
	}
	if(!g_imSpdir.isCached(buf)) { // Cache mishit
ods << "GetFileInfo - Cache mishit by " << buf << std::endl;
		yak::utils::windows::Slurp<CHAR> s(buf, len);
		SetArchiveInfo(buf, s.begin(), s.getSize());
	}
	g_imSpdir.getFileInfo(buf, filename, lpInfo);
	HANDLE hResult = CreateLinkOnMemory(g_imSpdir.getPIDL(buf, lpInfo->position), "created by afxspdir");
	lpInfo->compsize = lpInfo->filesize = LocalSize(hResult);
	LocalFree(hResult);
	return SPI_ERR_NO_ERROR;
}

INT PASCAL GetFile(LPSTR buf, LONG len, LPSTR dest, UINT flag, FARPROC prgressCallback, LONG lData)
{
//	ods << "GetFile(" << std::string(buf, std::min<DWORD>(len, 1024)) << ',' << len << ',' << ',' << flag << ',' << prgressCallback << ',' << lData << ')' << std::endl;
	ods << "GetFile(" << ',' << len << ',' << ',' << flag << ',' << prgressCallback << ',' << lData << ')' << std::endl;
	switch(flag & 7) {
	case 0:
	  {
		ods << "Source is file" << std::endl; // By afx
		break;
	  }
	case 1:
		ods << "Source is memory" << std::endl;
		return SPI_ERR_NOT_IMPLEMENTED; // We cannot handle this case.
	}
	switch((flag>>8) & 7) {
	case 0:
		ods << "Destination is file: " << dest << std::endl;
		return SPI_ERR_NOT_IMPLEMENTED; // We can but not yet implemented
	case 1:
		ods << "Destination is memory: " << reinterpret_cast<void*>(dest) << std::endl; // By afx
		break;
	}
	yak::utils::windows::Slurp<CHAR> s(buf, 0);
	if(!g_imSpdir.isCached(buf)) { // Cache mishit
ods << "GetFile - Cache mishit by " << buf << std::endl;
		SetArchiveInfo(buf, s.begin(), s.getSize());
	}
	HANDLE *pResult = reinterpret_cast<HANDLE*>(dest);
	*pResult = CreateLinkOnMemory(g_imSpdir.getPIDL(buf, len), "created by afxspdir");
	return SPI_ERR_NO_ERROR;
}
